// =============================================================================
//
//      --- kvi_malloc.cpp ---
//
//   This file is part of the KVIrc IRC client distribution
//   Copyright (C) 1999-2000 Szymon Stefanek (stefanek@tin.it)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
// =============================================================================

#define _KVI_MALLOC_CPP_

#include <stdio.h>

#include "kvi_malloc.h"
#include "kvi_settings.h"

#ifdef COMPILE_GCC_MEMORY_PROFILE

	//
	// Memory profile stuff
	// Used to find memory leaks etc...
	//

	typedef struct _KviMallocEntry
	{
		struct _KviMallocEntry *prev;
		void                   *pointer;
		int                     size;
		void                   *return_addr1;
		void                   *return_addr2;
		struct _KviMallocEntry *next;
	} KviMallocEntry;

	int             g_iMaxRequestSize           = 0;
	void           *g_pMaxRequestReturnAddress1 = 0;
	void           *g_pMaxRequestReturnAddress2 = 0;
	unsigned int    g_iMallocCalls              = 0;
	unsigned int    g_iReallocCalls             = 0;
	unsigned int    g_iFreeCalls                = 0;
	unsigned int    g_iTotalMemAllocated        = 0;
	unsigned int    g_uAllocationPeak           = 0;
	KviMallocEntry *g_pEntries                  = 0;

	void *kvi_malloc(int size)
	{
		g_iMallocCalls++;
		g_iTotalMemAllocated += size;
		if( g_iTotalMemAllocated > g_uAllocationPeak )
			g_uAllocationPeak = g_iTotalMemAllocated;
		if( g_iMaxRequestSize < size ) {
			g_iMaxRequestSize = size;
			g_pMaxRequestReturnAddress1 = __builtin_return_address(1);
			g_pMaxRequestReturnAddress2 = __builtin_return_address(2);
		}
		KviMallocEntry *e = (KviMallocEntry *) malloc(sizeof(KviMallocEntry));
		e->pointer = malloc(size);
		e->size = size;
		e->return_addr1 = __builtin_return_address(1);
		e->return_addr2 = __builtin_return_address(2);
		e->next = g_pEntries;
		e->prev = 0;
		if( g_pEntries ) g_pEntries->prev = e;
		g_pEntries = e;
		return e->pointer;
	}

	void *kvi_realloc(void *ptr, int size)
	{
		g_iReallocCalls++;
		if( !ptr ) return kvi_malloc(size);
		if( g_iMaxRequestSize < size ) {
			g_iMaxRequestSize = size;
			g_pMaxRequestReturnAddress1 = __builtin_return_address(1);
			g_pMaxRequestReturnAddress2 = __builtin_return_address(2);
		}
		KviMallocEntry *e = g_pEntries;
		while( e ) {
			if( e->pointer == ptr ) {
				g_iTotalMemAllocated -= e->size;
				g_iTotalMemAllocated += size;
				if( g_uAllocationPeak < g_iTotalMemAllocated )
					g_uAllocationPeak = g_iTotalMemAllocated;
				e->pointer = realloc(ptr, size);
				e->size = size;
				e->return_addr1 = __builtin_return_address(1);
				e->return_addr2 = __builtin_return_address(2);
				return e->pointer;
			}
			e = e->next;
		}
		fprintf(stderr, "Attempt to realloc a non-existant pointer (%p) (called from %p (%p))\n", ptr, __builtin_return_address(1), __builtin_return_address(2));
		return realloc(ptr, size);
	}

	void kvi_free(void *ptr)
	{
		g_iFreeCalls++;
		if( !ptr ) {
			fprintf(stderr, "Attempt to free a null pointer (called from %p (%p))\n", __builtin_return_address(1), __builtin_return_address(2));
			exit(-1);
		}
		KviMallocEntry *e = g_pEntries;
		while( e ) {
			if( e->pointer == ptr ) {
				g_iTotalMemAllocated -= e->size;
				if( e->prev ) {
					if( e == g_pEntries )
						fprintf(stderr, "Mem profiling internal error!\n");
					e->prev->next = e->next;
					if( e->next ) e->next->prev = e->prev;
				} else {
					if( e != g_pEntries )
						fprintf(stderr, "Mem profiling internal error!\n");
					if( e->next ) e->next->prev = 0;
					g_pEntries = e->next;
				}
				free(e);
				return;
			}
			e = e->next;
		}
		fprintf(stderr, "Attempt to free a nonexistant pointer (%p) (called from %p (%p))\n", ptr, __builtin_return_address(1), __builtin_return_address(2));
	}

	void kvi_memory_profile() __attribute__((destructor));
	void kvi_memory_profile()
	{
		unsigned int countUnfreed = 0;
		KviMallocEntry *e = g_pEntries;
		while( e ) {
			countUnfreed++;
			e = e->next;
		}
		fprintf(stderr, "|====|====|====|====|====|====|====|====\n");
		fprintf(stderr, "| Memory profile for KVIrc\n");
		fprintf(stderr, "| Unfreed chunks : %d\n", countUnfreed);
		fprintf(stderr, "| Total unfreed memory : %u bytes\n", g_iTotalMemAllocated);
		fprintf(stderr, "|====|====|====|====|====|====|====|====\n");
		fprintf(stderr, "| Possible unfreed chunks dump:\n");
		e = g_pEntries;
		while( e ) {
			fprintf(stderr, "|====|====|\n");
			fprintf(stderr, "| Currently unfreed chunk: %p\n", e->pointer);
			fprintf(stderr, "| Size: %d\n", e->size);
			fprintf(stderr, "| Caller address 1: %p\n", e->return_addr1);
			fprintf(stderr, "| Caller address 2: %p\n", e->return_addr2);
			fprintf(stderr, "| Data: \"%s\"\n", (char *) e->pointer);
			KviMallocEntry *toFree = e;
			e = e->next;
			free(toFree);
		}
		fprintf(stderr, "|====|====|====|====|====|====|====|====\n");
		fprintf(stderr, "| Allocation peak : %u bytes\n", g_uAllocationPeak);
		fprintf(stderr, "|====|====|====|====|====|====|====|====\n");
		fprintf(stderr, "| Max request size : %d bytes\n", g_iMaxRequestSize);
		fprintf(stderr, "| Called from %p (%p)\n", g_pMaxRequestReturnAddress1, g_pMaxRequestReturnAddress2);
		fprintf(stderr, "|====|====|====|====|====|====|====|====\n");
		fprintf(stderr, "| Malloc calls: %u\n", g_iMallocCalls);
		fprintf(stderr, "| Realloc calls: %u\n", g_iReallocCalls);
		fprintf(stderr, "| Free calls: %u\n", g_iFreeCalls);
		fprintf(stderr, "|====|====|====|====|====|====|====|====\n");
	}

#else // COMPILE_GCC_MEMORY_PROFILE

	#ifndef COMPILE_SKIP_MEMORY_CHECKS
		void outOfMemory()
		{
			fprintf(stderr, "Virtual memory exhausted in malloc call.\n");
			exit(-1);
		}
	#endif

#endif
