// =============================================================================
//
//      --- kvi_expression.cpp ---
//
//   This file is part of the KVIrc IRC client distribution
//   Copyright (C) 1999-2000 Szymon Stefanek (stefanek@tin.it)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
// =============================================================================

#define _KVI_DEBUG_CHECK_RANGE_
#define _KVI_DEBUG_CLASS_NAME_ "KviExprTree"

#include "kvi_command.h"
#include "kvi_debug.h"
#include "kvi_error.h"
#include "kvi_exprtree.h"
#include "kvi_locale.h"
#include "kvi_userparser.h"

long my_funky_pow(long x, long y)
{
	long total = 1;
	for( int i = 0; i < y; i++ )
		total = total*x;
	return total;
}

KviExprTree::KviExprTree(NodeType t, long v)
{
	m_type   = t;
	m_left   = 0;
	m_right  = 0;
	m_value  = v;
	m_parent = 0;
}

KviExprTree::~KviExprTree()
{
	if( m_left  ) { delete m_left;  m_left  = 0; }
	if( m_right ) { delete m_right; m_right = 0; }
}

KviExprTree *KviExprTree::left()
{
	return m_left;
}

KviExprTree *KviExprTree::right()
{
	return m_right;
}

void KviExprTree::setLeft(KviExprTree *node)
{
	__range_valid(node);
	m_left = node;
	node->m_parent = this;
}

void KviExprTree::setRight(KviExprTree *node)
{
	__range_valid(node);
	m_right = node;
	node->m_parent = this;
}

long KviExprTree::calculate(KviCommand *cmd)
{
	if( m_type == NumericOperand )
		return m_value;

	if( m_type == StringOperand ) {
		// Possible only if this is the only operand in the sub-tree,
		// has no operator, and no second operand.
		cmd->setError(KVI_ERROR_SingleStringOperand, _i18n_("EXPRESSION-CORE"), m_string.ptr());
		return -1;
	}
	if( cmd->hasError() )
		return -1;

	__range_valid(m_left);
	if( m_type == UnaryOperator ) {
		// Valid only for numeric operands
		if( m_left->m_type == StringOperand ) {
			cmd->setError(KVI_ERROR_OperandDoesNotMatchOperator, _i18n_("EXPRESSION-CORE"), m_left->m_string.ptr());
			return -1;
		}
		switch( m_value ) {
			case OP_NOT:    return !m_left->calculate(cmd); break;
			case OP_NEGATE: return -m_left->calculate(cmd); break;
			case OP_BITNOT: return ~m_left->calculate(cmd); break;
			default:
				__range_invalid(true); // Error
			break;
		}
	}

	__range_valid(m_right);
	if( (m_value < 10) || (m_value > 16) ) {
		// Numeric operators...
		if( (m_left->m_type == StringOperand) || (m_right->m_type == StringOperand) ) {
			cmd->setError(KVI_ERROR_OperandDoesNotMatchOperator, _i18n_("EXPRESSION-CORE"), m_left->m_string.ptr());
			return -1;
		}
		switch( m_value ) {
			case OP_BITAND:
				return (m_left->calculate(cmd) & m_right->calculate(cmd));
				break;
			case OP_BITOR:
				return (m_left->calculate(cmd) | m_right->calculate(cmd));
				break;
			case OP_SHR:
				return (m_left->calculate(cmd) >> m_right->calculate(cmd));
				break;
			case OP_SHL:
				return (m_left->calculate(cmd) << m_right->calculate(cmd));
				break;
			case OP_POWER:
				return my_funky_pow(m_left->calculate(cmd), m_right->calculate(cmd));
				break;
			case OP_MUL:
				return (m_left->calculate(cmd) * m_right->calculate(cmd));
				break;
			case OP_DIV: {
				long temp = m_right->calculate(cmd);
				if( temp == 0 ) {
					cmd->setError(KVI_ERROR_DivisionByZero, _i18n_("EXPRESSION-CORE"));
					return -1;
				}
				return (m_left->calculate(cmd) / temp);
				break;
			}
			case OP_MOD: {
				long temp = m_right->calculate(cmd);
				if( temp == 0 ) {
					cmd->setError(KVI_ERROR_DivisionByZero, _i18n_("EXPRESSION-CORE"));
					return -1;
				}
				return (m_left->calculate(cmd) % temp);
				break;
			}
			case OP_ADD:
				return (m_left->calculate(cmd) + m_right->calculate(cmd));
				break;
			case OP_SUB:
				return (m_left->calculate(cmd) - m_right->calculate(cmd));
				break;
			case OP_AND:
				return (m_left->calculate(cmd) && m_right->calculate(cmd));
				break;
			case OP_OR:
				return (m_left->calculate(cmd) || m_right->calculate(cmd));
				break;
			default:
				__range_invalid(true); // Tragedy :)
				break;
		}
	}

	if( m_left->m_type == StringOperand ) {
		if( m_right->m_type != StringOperand ) {
			cmd->setError(KVI_ERROR_OperandDoesNotMatchOperator, _i18n_("EXPRESSION-CORE"), m_left->m_string.ptr());
			return -1;
		}
		// String operator for string operands
		switch( m_value ) {
			case OP_GT:
				return (qstricmp(m_left->m_string.ptr(), m_right->m_string.ptr()) < 0);
				break;
			case OP_LT:
				return (qstricmp(m_left->m_string.ptr(), m_right->m_string.ptr()) > 0);
				break;
			case OP_EQ:
				return (qstricmp(m_left->m_string.ptr(), m_right->m_string.ptr()) == 0);
				break;
			case OP_CSEQ:
				return (qstrcmp(m_left->m_string.ptr(), m_right->m_string.ptr()) == 0);
				break;
			case OP_GE:
				return (qstricmp(m_left->m_string.ptr(), m_right->m_string.ptr()) <= 0);
				break;
			case OP_LE:
				return (qstricmp(m_left->m_string.ptr(), m_right->m_string.ptr()) >= 0);
				break;
			case OP_NE:
				return (qstrcmp(m_left->m_string.ptr(), m_right->m_string.ptr()) != 0);
				break;
			default:
				__range_invalid(true);
				break;
		}
	} else if( m_right->m_type == StringOperand ) {
		cmd->setError(KVI_ERROR_OperandDoesNotMatchOperator, _i18n_("EXPRESSION-CORE"), m_right->m_string.ptr());
		return -1;
	}

	// Numeric operator for numeric operands
	switch( m_value ) {
		case OP_GT:
			return (m_left->calculate(cmd) > m_right->calculate(cmd));
			break;
		case OP_LT:
			return (m_left->calculate(cmd) < m_right->calculate(cmd));
			break;
		case OP_EQ:
			return (m_left->calculate(cmd) == m_right->calculate(cmd));
			break;
		case OP_CSEQ:
			cmd->setError(KVI_ERROR_OperandDoesNotMatchOperator, _i18n_("EXPRESSION-CORE"));
			return -1;
			break;
		case OP_GE:
			return (m_left->calculate(cmd) >= m_right->calculate(cmd));
			break;
		case OP_LE:
			return (m_left->calculate(cmd) <= m_right->calculate(cmd));
			break;
		case OP_NE:
			return (m_left->calculate(cmd) != m_right->calculate(cmd));
			break;
		default:
			__range_invalid(true);
			break;
	}
	return -1; // Never here...
}

KviExprTree *KviExprTree::parentWithPrecedenceLowerThan(long operValue)
{
	__range_valid(m_type == BinaryOperator);
	if( m_value >= operValue )
		return this;
	if( !m_parent )
		return 0;
	return m_parent->parentWithPrecedenceLowerThan(operValue);
}

long KviExprTree::firstBinaryOperator()
{
	if( m_type == BinaryOperator )
		return m_value;
	__range_valid(m_type == UnaryOperator);
	return m_left->firstBinaryOperator();
}
