// =============================================================================
//
//      --- kvi_userpopupmenu.cpp ---
//
//   This file is part of the KVIrc IRC client distribution
//   Copyright (C) 1999-2000 Szymon Stefanek (stefanek@tin.it)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
// =============================================================================

#define _KVI_DEBUG_CHECK_RANGE_
#define _KVI_DEBUG_CLASS_NAME_ "KviUserPopupMenu"

#include <qcursor.h>
#include <qfile.h>

#include "kvi_app.h"
#include "kvi_command.h"
#include "kvi_debug.h"
#include "kvi_defines.h"
#include "kvi_error.h"
#include "kvi_fileutils.h"
#include "kvi_frame.h"
#include "kvi_locale.h"
#include "kvi_userpopupmenu.h"
#include "kvi_userparser.h"
#include "kvi_window.h"

// Declared in kvi_app.cpp
extern QPixmap          *g_pixViewOut[KVI_OUT_NUM_IMAGES];
extern KviUserPopupMenu *g_pChannelPopup;
extern KviUserPopupMenu *g_pConsolePopup;
extern KviUserPopupMenu *g_pUserlistPopup;
extern KviUserPopupMenu *g_pNotifylistPopup;
extern KviUserPopupMenu *g_pQueryPopup;
extern KviUserPopupMenu *g_pDccChatPopup;

KviUserPopupMenu::KviUserPopupMenu(QWidget *parent, const char *name)
	: KviPopupMenu(parent, name)
{
	m_pDataList.setAutoDelete(true);
	connect(this, SIGNAL(activated(int)), this, SLOT(localItemClicked(int)));
	m_pParent       = 0;
	m_pParentWindow = 0;
	m_bInTestMode   = false;
	m_bLocked       = false;
}

KviUserPopupMenu::~KviUserPopupMenu()
{
	clearAll();
}

void KviUserPopupMenu::copyFromPopup(KviUserPopupMenu *popup)
{
	__range_valid(popup);
	clearAll();
	int idx = 0;
	for( KviUserPopupMenuData *d = popup->m_pDataList.first(); d; d = popup->m_pDataList.next() ) {
		if( d->menu ) {
			KviStr yourName(KviStr::Format, "%s_sub%d", name(), idx);
			KviUserPopupMenu *sub = new KviUserPopupMenu(0, yourName.ptr());
			sub->copyFromPopup(d->menu);
			insertPopupItem(d->text.ptr(), d->image.ptr(), sub);
			idx++;
		} else {
			if( (d->id == 0) && kvi_strEqualCS(d->buffer.ptr(), " ") ) {
				insertSeparatorItem();
			} else {
				insertNormalItem(d->text.ptr(), d->image.ptr(), d->buffer.ptr());
			}
		}
	}
}

void KviUserPopupMenu::insertSeparatorItem()
{
	KviUserPopupMenuData *data = new KviUserPopupMenuData();
	insertSeparator();
	data->id     = 0;
	data->buffer = " ";
	data->text   = "";
	data->menu   = 0;
	m_pDataList.append(data);
}

int KviUserPopupMenu::insertNormalItem(const char *name, const char *img, const char *buffer)
{
	KviUserPopupMenuData *data = new KviUserPopupMenuData();
	KviStr pName(name);
	if( m_pParentWindow && m_pParentWindow->frame() )
		m_pParentWindow->frame()->m_pUserParser->processString(name, m_pParentWindow, pName);
	KviStr image(img);
	if( !image.hasData() )
		image = "0"; // Always have an image visible
	if( image.isNum() ) {
		bool bOk = false;
		int imId = image.toInt(&bOk);
		if( !bOk || (imId < 0) )
			imId = 0;
		else
			imId = imId % KVI_OUT_NUM_IMAGES;
		data->id = insertItem(QIconSet(*g_pixViewOut[imId]), _CHAR_2_QSTRING(pName.ptr()));
	} else {
		KviStr imagePath;
		data->id = -1;
		if( g_pApp->findImage(imagePath, image.ptr()) ) {
			QPixmap pix(imagePath.ptr());
			if( !pix.isNull() )
				data->id = insertItem(QIconSet(pix), _CHAR_2_QSTRING(pName.ptr()));
		}
		if( data->id == -1 )
			data->id = insertItem(QIconSet(*g_pixViewOut[0]), _CHAR_2_QSTRING(pName.ptr()));
	}
	data->buffer = buffer;
	data->image  = image;
	data->menu   = 0;
	data->text   = name;
	data->buffer.stripWhiteSpace();
	m_pDataList.append(data);
	return data->id;
}

void KviUserPopupMenu::insertPopupItem(const char *name, const char *img, KviUserPopupMenu *menu)
{
	KviUserPopupMenuData *data = new KviUserPopupMenuData();
	KviStr pName(name);
	if( m_pParentWindow && m_pParentWindow->frame() )
		m_pParentWindow->frame()->m_pUserParser->processString(name, m_pParentWindow, pName);
	KviStr image(img);
	if( !image.hasData() )
		image = "0";
	if( image.isNum() ) {
		bool bOk = false;
		int imId = image.toInt(&bOk);
		if( !bOk || (imId < 0) )
			imId = 0;
		else
			imId = imId % KVI_OUT_NUM_IMAGES;
		data->id = insertItem(QIconSet(*g_pixViewOut[imId]), _CHAR_2_QSTRING(pName.ptr()), menu);
	} else {
		KviStr imagePath;
		data->id = -1;
		if( g_pApp->findImage(imagePath, image.ptr()) ) {
			QPixmap pix(imagePath.ptr());
			if( !pix.isNull() )
				data->id = insertItem(QIconSet(pix), _CHAR_2_QSTRING(pName.ptr()), menu);
		}
		if( data->id == -1 )
			data->id = insertItem(QIconSet(*g_pixViewOut[0]), _CHAR_2_QSTRING(pName.ptr()), menu);
	}
	data->image = image;
	data->menu  = menu;
	data->text  = name;
	m_pDataList.append(data);
	menu->m_pParent       = this;
	menu->m_pParentWindow = m_pParentWindow;
}

void KviUserPopupMenu::localItemClicked(int id)
{
	for( KviUserPopupMenuData *d = m_pDataList.first(); d; d = m_pDataList.next() ) {
		if( d->id == id && !(d->menu) ) {
			// Only the top menu emits itemClicked() -- or testItemClicked if in test mode
			if( m_pParent )
				m_pParent->itemClickedRedirect(d->buffer, d->id);
			else {
				if( m_bInTestMode )
					emit testItemClicked(d->id);
				else {
					m_bLocked = true;
					emit itemClicked(d->buffer);
					m_bLocked = false;
				}
			}
			return;
		}
	}
	__debug_1arg("Oops... cannot find menu item %d", id);
}

void KviUserPopupMenu::itemClickedRedirect(const KviStr &buffer, int id)
{
	if( m_pParent )
		m_pParent->itemClickedRedirect(buffer, id);
	else {
		if( m_bInTestMode )
			emit testItemClicked(id);
		else {
			m_bLocked = true;
			emit itemClicked(buffer);
			m_bLocked = false;
		}
	}
}

void KviUserPopupMenu::clearAll()
{
	KviUserPopupMenuData *data;
	while( (data = m_pDataList.first()) ) {
		if( data->menu )
			delete data->menu;
		m_pDataList.removeFirst();
	}
	clear();
}

const KviStr KviUserPopupMenu::text(int id)
{
	KviUserPopupMenuData *d;
	for( d = m_pDataList.first(); d; d = m_pDataList.next() ) {
		if( d->id == id )
			return d->text;
	}
	const KviStr empty;
	return empty;
}

void KviUserPopupMenu::setFont(const QFont &font)
{
	// Propagate the font to the children
	for( KviUserPopupMenuData *d = m_pDataList.first(); d; d = m_pDataList.next() ) {
		if( d->menu )
			d->menu->setFont(font);
	}
	KviPopupMenu::setFont(font);
}

void KviUserPopupMenu::setPaletteRecursive(const QPalette &pal)
{
	// Propagate the palette to the children
	for( KviUserPopupMenuData *d = m_pDataList.first(); d; d = m_pDataList.next() ) {
		if( d->menu )
			d->menu->setPalette(pal);
	}
	setPalette(pal);
}

void KviUserPopupMenu::doPopup(const QWidget *receiver, const char *memberSlot)
{
	disconnect(this, SIGNAL(itemClicked(const KviStr &)), 0, 0);
	if( m_bInTestMode ) {
		disconnect(this, SIGNAL(testItemClicked(int)), 0, 0);
		m_bInTestMode = false;
	}

	connect(this, SIGNAL(itemClicked(const KviStr &)), receiver, memberSlot);
	reloadMenu(receiver);
	popup(QCursor::pos());
}

/**
 * Overloaded for fixed pos
 */
void KviUserPopupMenu::doPopup(const QWidget *receiver, const char *memberSlot, QPoint pos)
{
	disconnect(this, SIGNAL(itemClicked(const KviStr &)), 0, 0);
	if( m_bInTestMode ) {
		disconnect(this, SIGNAL(testItemClicked(int)), 0, 0);
		m_bInTestMode = false;
	}

	connect(this, SIGNAL(itemClicked(const KviStr &)), receiver, memberSlot);
	reloadMenu(receiver);
	popup(pos);
}

void KviUserPopupMenu::doTestPopup(const QWidget *receiver, const char *memberSlot)
{
	disconnect(this, SIGNAL(itemClicked(const KviStr &)), 0, 0);
	disconnect(this, SIGNAL(testItemClicked(int)), 0, 0);
	m_bInTestMode = true;

	connect(this, SIGNAL(testItemClicked(int)), receiver, memberSlot);
	reloadMenu(receiver);
	popup(QCursor::pos());
}

void KviUserPopupMenu::reloadMenu(const QWidget *parentWindow)
{
	KviStr buffer;
	KviStr error;
	m_pParentWindow = (KviWindow *) parentWindow;
	if( !m_pParentWindow )
		m_pParentWindow = (KviWindow *) g_pApp->m_pFrameList->first()->m_pConsole;
	buildDataBuffer(buffer);
	loadFromDataBuffer(buffer, error);
}

/**
 * ===========================================================================
 *
 * Load and save popups.
 *
 */
bool KviUserPopupMenu::saveAllPopups(const char *filename)
{
	// TODO: Maybe write a simpler code to save and load popups?
	QFile f(filename);
	if( !f.open(IO_WriteOnly | IO_Truncate) )
		return false;

	f.writeBlock(KVI_MAGIC_STRING, KVI_MAGIC_STRING_LENGTH);

	g_pChannelPopup->save   (&f, "channel");
	g_pConsolePopup->save   (&f, "console");
	g_pUserlistPopup->save  (&f, "userlist");
	g_pNotifylistPopup->save(&f, "notifylist");
	g_pQueryPopup->save     (&f, "query");
	g_pDccChatPopup->save   (&f, "dccchat");

	f.close();
	return true;
}

bool KviUserPopupMenu::loadAllPopups(const char *filename, KviStr &error)
{
	KviStr buffer;
	if( !kvi_loadFile(filename, buffer) ) {
		error = _i18n_("Unable to load file");
		return false;
	}

	if( !kvi_strEqualCSN(KVI_MAGIC_STRING, buffer.ptr(), KVI_MAGIC_STRING_LENGTH) ) {
		error = _i18n_("Invalid magic");
		__debug(buffer.ptr());
		return false; // Not a KVIrc file
	}

	buffer.cutLeft(KVI_MAGIC_STRING_LENGTH);

	while( buffer.hasData() ) {
		buffer.stripLeftWhiteSpace();

		if( !kvi_strEqualCSN("[POPUP ", buffer.ptr(), 7) ) {
			error = _i18n_("Invalid POPUP label");
			__debug(buffer.ptr());
			return false;
		}

		buffer.cutLeft(7);
		buffer.stripLeftWhiteSpace();

		KviUserPopupMenu *target = 0;

		if( kvi_strEqualNoLocaleCIN("channel", buffer.ptr(), 7) ) {
			target = g_pChannelPopup;
			buffer.cutLeft(7);
		} else if( kvi_strEqualNoLocaleCIN("console", buffer.ptr(), 7) ) {
			target = g_pConsolePopup;
			buffer.cutLeft(7);
		} else if( kvi_strEqualNoLocaleCIN("userlist", buffer.ptr(), 8) ) {
			target = g_pUserlistPopup;
			buffer.cutLeft(8);
		} else if( kvi_strEqualNoLocaleCIN("notifylist", buffer.ptr(), 10) ) {
			target = g_pNotifylistPopup;
			buffer.cutLeft(10);
		} else if( kvi_strEqualNoLocaleCIN("query", buffer.ptr(), 5) ) {
			target = g_pQueryPopup;
			buffer.cutLeft(5);
		} else if( kvi_strEqualNoLocaleCIN("dccchat", buffer.ptr(), 7) ) {
			target = g_pDccChatPopup;
			buffer.cutLeft(7);
		} else {
			error = __tr("Unknown popup name");
			__debug(buffer.ptr());
			return false;
		}

		buffer.stripLeftWhiteSpace();
		if( !kvi_strEqualNoLocaleCIN("bytes ", buffer.ptr(), 6) ) {
			error = __tr("Syntax error in POPUP label");
			__debug(buffer.ptr());
			return false;
		}

		buffer.cutLeft(6);
		buffer.stripLeftWhiteSpace();

		if( !isdigit(*(buffer.ptr())) ) {
			error = _i18n_("Syntax error in [POPUP] label: digit expected");
			__debug(buffer.ptr());
			return false;
		}

		KviStr numBytes;
		if( !buffer.getToken(numBytes, ']') ) {
			error = _i18n_("Syntax error in [POPUP] label: cannot extract popup length");
			__debug(buffer.ptr());
			return false;
		}

		bool bOk = false;
		int bytes = numBytes.toInt(&bOk);
		if( !bOk ) {
			error  = _i18n_("Unrecogized token ");
			error += numBytes;
			__debug(buffer.ptr());
			return false;
		} // Syntax error again... someone messed with the file

		buffer.cutLeft(1); // Cut the newline
		if( buffer.len() < bytes ) {
			error = _i18n_("Popup buffer smaller than declared length (");
			error += numBytes;
			error += ")";
			__debug(buffer.ptr());
			return false;
		}

		KviStr popupBuf = buffer.left(bytes);
		buffer.cutLeft(bytes);
		buffer.stripLeftWhiteSpace();

		if( !kvi_strEqualCSN("[ENDPOPUP]\n", buffer.ptr(), 11) ) {
			error = __tr("Something went wrong with the popup buffer length");
			__debug(buffer.ptr());
			return false;
		}
		buffer.cutLeft(11);

		if( target->m_bLocked ) {
			error = __tr("Attempted to modify a locked popup");
			debug("WARNING: attempted to modify locked popup, skipping");
			return false;
		}

		if( !target->loadFromDataBuffer(popupBuf, error) ) {
			__debug(popupBuf.ptr());
			return false;
		}
	}
	return true;
}

void KviUserPopupMenu::save(QFile *f, const char *popupName)
{
	f->writeBlock("[POPUP ", 7);
	f->writeBlock(popupName, strlen(popupName));
	f->writeBlock(" bytes ", 7);

	KviStr buffer;
	buildDataBuffer(buffer);

	KviStr szNBytes(KviStr::Format, "%d", buffer.len());
	f->writeBlock(szNBytes.ptr(), szNBytes.len());
	f->writeBlock("]\n", 2);
	f->writeBlock(buffer.ptr(), buffer.len());
	f->writeBlock("\n[ENDPOPUP]\n", 12);
}

/**
 * An item with id = 0, menu = 0 and buffer = " " is a separator.
 * Normal menu item buffers have the initial spaces stripped at
 * insertion time.
 */
bool KviUserPopupMenu::loadFromDataBuffer(KviStr &buffer, KviStr &error)
{
	clearAll();
	buffer.stripLeftWhiteSpace();
	int idx  = 0;
	while( buffer.hasData() ) {
		if( kvi_strEqualNoLocaleCIN("[separator]\n", buffer.ptr(), 12) ) {
			buffer.cutLeft(12);
			insertSeparatorItem();
		} else if( kvi_strEqualNoLocaleCIN("[item]\n", buffer.ptr(), 7) ) {
			buffer.cutLeft(7);
			KviStr image, name, bytes;
			if( !buffer.getLine(image) ) {
				error = __tr("Unexpected end of buffer");
				return false;
			}
			buffer.stripLeftWhiteSpace();
			if( !buffer.getLine(name) ) {
				error = __tr("Unexpected end of buffer");
				return false;
			}
			buffer.stripLeftWhiteSpace();
			if( !buffer.getLine(bytes) ) {
				error = __tr("Unexpected end of buffer");
				return false;
			}

			if( !kvi_strEqualNoLocaleCIN("image=", image.ptr(), 6) ) {
				error = __tr("Unrecognized item image label");
				return false;
			}
			image.cutLeft(6);
			image.stripWhiteSpace();
			if( !kvi_strEqualNoLocaleCIN("name=", name.ptr(), 5) ) {
				error = __tr("Unrecognized item name label");
				return false;
			}
			name.cutLeft(5);
			name.stripWhiteSpace();
			if( !kvi_strEqualNoLocaleCIN("bytes=", bytes.ptr(), 6) ) {
				error = __tr("Unrecognized item length label");
				return false;
			}
			bytes.cutLeft(6);
			bytes.stripWhiteSpace();
			bool bOk = false;
			int bLen = bytes.toInt(&bOk);
			if( !bOk ) {
				error = __tr("Unrecognized item length");
				return false;
			}
			if( (bLen < 0) || (bLen > buffer.len()) ) {
				error = __tr("Buffer is shorter than the declared length");
				return false;
			}
			KviStr pBuf = buffer.left(bLen);
			buffer.cutLeft(bLen);
			buffer.stripLeftWhiteSpace();
			if( !kvi_strEqualNoLocaleCIN("[enditem]\n", buffer.ptr(), 10) ) {
				error = __tr("Something went wrong with the buffer length");
				return false;
			}
			buffer.cutLeft(10);
			insertNormalItem(name.ptr(), image.ptr(), pBuf.ptr());
		} else if( kvi_strEqualNoLocaleCIN("[submenu]\n", buffer.ptr(), 10) ) {
			buffer.cutLeft(10);
			KviStr image, szName, bytes;
			if( !buffer.getLine(image) ) {
				error = __tr("Unexpected end of buffer");
				return false;
			}
			buffer.stripLeftWhiteSpace();
			if( !buffer.getLine(szName) ) {
				error = __tr("Unexpected end of buffer");
				return false;
			}
			buffer.stripLeftWhiteSpace();
			if( !buffer.getLine(bytes) ) {
				error = __tr("Unexpected end of buffer");
				return false;
			}

			if( !kvi_strEqualNoLocaleCIN("image=", image.ptr(), 6) ) {
				error = __tr("Unrecognized submenu image label");
				return false;
			}
			image.cutLeft(6);
			image.stripWhiteSpace();
			if( !kvi_strEqualNoLocaleCIN("name=", szName.ptr(), 5) ) {
				error = __tr("Unrecognized submenu name label");
				return false;
			}
			szName.cutLeft(5);
			szName.stripWhiteSpace();
			if( !kvi_strEqualNoLocaleCIN("bytes=", bytes.ptr(), 6) ) {
				error = __tr("Unrecognized submenu length label");
				return false;
			}
			bytes.cutLeft(6);
			bytes.stripWhiteSpace();
			bool bOk = false;
			int bLen = bytes.toInt(&bOk);
			if( !bOk ) {
				error = __tr("Unrecognized submenu length");
				return false;
			}
			if( (bLen < 0) || (bLen > buffer.len()) ) {
				error = __tr("Buffer is shorter than the declared length");
				return false;
			}
			KviStr mBuf = buffer.left(bLen);
			buffer.cutLeft(bLen);
			buffer.stripLeftWhiteSpace();
			if( !kvi_strEqualNoLocaleCIN("[endsubmenu]\n", buffer.ptr(), 13) ) {
				error = __tr("Something went wrong with the buffer length");
				return false;
			}
			buffer.cutLeft(13);

			KviStr yourName(KviStr::Format, "%s_sub%d", name(), idx);
			KviUserPopupMenu *menu = new KviUserPopupMenu(0, yourName.ptr());
			if( !menu->loadFromDataBuffer(mBuf, error) ) {
				delete menu;
				return false;
			}

			insertPopupItem(szName.ptr(), image.ptr(), menu);
			idx++;
		} else {
			error = __tr("Unrecognized popup item");
			return false;
		}
		buffer.stripLeftWhiteSpace();
	}
	if( count() > 0 )
		insertTearOffHandle();
	return true;
}

void KviUserPopupMenu::buildDataBuffer(KviStr &str)
{
	for( KviUserPopupMenuData *d = m_pDataList.first(); d; d = m_pDataList.next() ) {
		if( !(d->menu) ) {
			if( d->id == 0 ) {
				if( kvi_strEqualCSN(d->buffer.ptr(), " ", 1) ) {
					str.append("[separator]\n");
					continue;
				}
			}
			str.append("[item]\n");
			str.append("image=");
			if( d->image.hasData() )
				str.append(d->image);
			str.append("\nname=");
			str.append(d->text);
			str.append("\nbytes=");
			KviStr tmp(KviStr::Format, "%d", d->buffer.len());
			str.append(tmp);
			str.append("\n");
			str.append(d->buffer);
			str.append("\n[enditem]\n");
		} else {
			str.append("[submenu]\n");
			str.append("image=");
			if( d->image.hasData() )
				str.append(d->image);
			str.append("\nname=");
			str.append(d->text);
			KviStr auxBuf;
			d->menu->buildDataBuffer(auxBuf);
			str.append("\nbytes=");
			KviStr tmp(KviStr::Format, "%d", auxBuf.len());
			str.append(tmp);
			str.append("\n");
			str.append(auxBuf);
			str.append("\n[endsubmenu]\n");
		}
	}
}

/**
 * =============================================================================
 *
 * Uparser code to generate menus: auxiliary function
 *
 * popup(popupname) <-- uparser
 * { <---- This function starts here
 * } <---- And ends here
 *
 */
bool KviUserParser::generatePopupFromBuffer(KviUserPopupMenu *m, KviCommand *cmd, bool bAppend)
{
	if( !bAppend )
		m->clearAll();

	if( *(cmd->m_ptr) != '{' ) {
		cmd->setError(KVI_ERROR_OpenBraceExpected, "POPUP-GENERATOR", cmd->m_ptr);
		return false;
	}
	++(cmd->m_ptr);

	int idx = 0;

	for( ;; ) {
		cmd->skipWhitespace();
		if( !(*(cmd->m_ptr)) ) { // Missing terminator
			cmd->setError(KVI_ERROR_UnexpectedEndOfCommand, "POPUP-GENERATOR");
			return false;
		}
		if( *(cmd->m_ptr) == '}' ) { // Finished
			++(cmd->m_ptr);
			return true;
		}

		if( kvi_strEqualNoLocaleCIN(cmd->m_ptr, "item", 4) ) {
			cmd->m_ptr += 4;
			cmd->skipSpace();
			if( *(cmd->m_ptr) != '(' ) {
				cmd->setError(KVI_ERROR_OpenParenthesisExpected, "POPUP-ITEM-LABEL", cmd->m_ptr);
				return false;
			}
			++(cmd->m_ptr);
			KviStr itemLabel;
			if( !processFncSingleParam(cmd, itemLabel) )
				return false;

			KviStr itemImage;
			if( !processFncFinalPart(cmd, itemImage) )
				return false;
			cmd->skipSpace();

			if( *(cmd->m_ptr) == ':' ) { // Allowed but not required
				++(cmd->m_ptr);
				cmd->skipSpace();
			}

			if( !(*(cmd->m_ptr)) ) {
				cmd->setError(KVI_ERROR_PopupItemCommandExpected, "POPUP-ITEM-LABEL");
				return false;
			}

			const char *aux = cmd->m_ptr;
			if( !skipCommand(cmd) )
				return false;
			KviStr buffer(aux, cmd->m_ptr);
			m->insertNormalItem(itemLabel.ptr(), itemImage.ptr(), buffer.ptr());
			continue;
		}

		if( kvi_strEqualNoLocaleCIN("separator", cmd->m_ptr, 9) ) {
			cmd->m_ptr += 9;
			cmd->skipSpace();
			if( (*(cmd->m_ptr) == ':') || (*(cmd->m_ptr) == ';') ) { // Allowed but not required
				++(cmd->m_ptr);
			}
			m->insertSeparatorItem();
			continue;
		}
		if( kvi_strEqualNoLocaleCIN(cmd->m_ptr, "submenu", 7) ) {
			cmd->m_ptr += 7;
			cmd->skipSpace();
			if( *(cmd->m_ptr) != '(' ) {
				cmd->setError(KVI_ERROR_OpenParenthesisExpected, "POPUP-SUBMENU-LABEL", cmd->m_ptr);
				return false;
			}
			++(cmd->m_ptr);
			KviStr itemLabel;
			if( !processFncSingleParam(cmd, itemLabel) )
				return false;

			KviStr itemImage;
			if( !processFncFinalPart(cmd, itemImage) )
				return false;
			cmd->skipSpace();

			if( *(cmd->m_ptr) == ':' ) { // Allowed but not required
				++(cmd->m_ptr);
				cmd->skipSpace();
			}

			if( !(*(cmd->m_ptr)) ) {
				cmd->setError(KVI_ERROR_PopupItemCommandExpected, "POPUP-SUBMENU-LABEL");
				return false;
			}

			KviStr yourName(KviStr::Format, "%s_sub%d", m->name(), idx);
			KviUserPopupMenu *menu = new KviUserPopupMenu(0, yourName.ptr());
			if( !generatePopupFromBuffer(menu, cmd, false) ) {
				delete menu;
				return false;
			}
			m->insertPopupItem(itemLabel.ptr(), itemImage.ptr(), menu);
			idx++;
			continue;
		}
		cmd->setError(KVI_ERROR_ItemPopupOrSeparatorExpected, "POPUP-GENERATOR", cmd->m_ptr);
		return false;
	}
}

// TODO: Set all KVIPOPUP MENU parents to 0!!

bool KviUserParser::skipPopupBuffer(KviCommand *cmd)
{
	if( *(cmd->m_ptr) != '{' ) {
		cmd->setError(KVI_ERROR_OpenBraceExpected, "POPUP-GENERATOR", cmd->m_ptr);
		return false;
	}
	++(cmd->m_ptr);
	for( ;; ) {
		cmd->skipWhitespace();
		if( !(*(cmd->m_ptr)) ) {
			cmd->setError(KVI_ERROR_UnexpectedEndOfCommand, "POPUP-GENERATOR");
			return false;
		}
		if( *(cmd->m_ptr) == '}' ) {
			++(cmd->m_ptr);
			return true;
		}
		if( kvi_strEqualNoLocaleCIN(cmd->m_ptr, "item", 4) ) {
			cmd->m_ptr += 4;
			cmd->skipSpace();
			if( *(cmd->m_ptr) != '(' ) {
				cmd->setError(KVI_ERROR_OpenParenthesisExpected, "POPUP-ITEM-LABEL", cmd->m_ptr);
				return false;
			}
			++(cmd->m_ptr);
			if( !skipExpressionBody(cmd) )
				return false;
			cmd->skipSpace();
			if( *(cmd->m_ptr) == ':' ) {
				++(cmd->m_ptr);
				cmd->skipSpace();
			}
			if( !(*(cmd->m_ptr)) ) {
				cmd->setError(KVI_ERROR_PopupItemCommandExpected, "POPUP-ITEM-LABEL");
				return false;
			}
			if( !skipCommand(cmd) )
				return false;
			continue;
		}
		if( kvi_strEqualNoLocaleCIN("separator", cmd->m_ptr, 9) ) {
			cmd->m_ptr += 9;
			cmd->skipSpace();
			if( (*(cmd->m_ptr) == ':') || (*(cmd->m_ptr) == ';') ) {
				++(cmd->m_ptr);
			}
			continue;
		}
		if( kvi_strEqualNoLocaleCIN(cmd->m_ptr, "submenu", 7) ) {
			cmd->m_ptr += 7;
			cmd->skipSpace();
			if( *(cmd->m_ptr) != '(' ) {
				cmd->setError(KVI_ERROR_OpenParenthesisExpected, "POPUP-SUBMENU-LABEL", cmd->m_ptr);
				return false;
			}
			++(cmd->m_ptr);
			if( !skipExpressionBody(cmd) )
				return false;
			cmd->skipSpace();
			if( *(cmd->m_ptr) == ':' ) {
				++(cmd->m_ptr);
				cmd->skipSpace();
			}
			if( !(*(cmd->m_ptr)) ) {
				cmd->setError(KVI_ERROR_PopupItemCommandExpected, "POPUP-SUBMENU-LABEL");
				return false;
			}
			if( !skipPopupBuffer(cmd) )
				return false;
			continue;
		}
		cmd->setError(KVI_ERROR_ItemPopupOrSeparatorExpected, "POPUP-GENERATOR", cmd->m_ptr);
		return false;
	}
}

#include "m_kvi_userpopupmenu.moc"
